import socket
import sys
from common import *
from buffer import FileBuffer
import time

class Client():

    def __init__(self, port=10001, filepath="out"):
        self.PORT = port
        self.FILEPATH = filepath

        self.init_socket()
        self.broadcast()
        try:
            self.handshake()
            self.recv_file()
            self.save_file()
        except ConnectionError:
            log("!", "Connecion error.")
            pass

    def init_socket(self):
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try :
            if self.PORT == BROADCAST_PORT:
                raise ValueError
            self.socket.bind((HOST_ADDRESS, self.PORT))                
        except :
            log("!", "Error when assigning port, continue using random port")
            pass
    
    def send(self, data, dst_addr):
        self.socket.sendto(data, dst_addr)

    def recv(self):
        return self.socket.recvfrom(SEG_SIZE)

    def broadcast(self):
        self.send(bytes(), (HOST_ADDRESS, BROADCAST_PORT))
        log("!", "Message sent to broadcast address")
        log("!", "Listening for three-way handshake...")

    def handshake(self):
        rcvd = Segment()
        while 1:
            byte, server_addr = self.recv()
            rcvd.from_bytes(byte)
            if not rcvd.test_checksum():
                log("!", "Checksum failed.")
                raise ConnectionError
            if rcvd.getSYN():
                break;

        self.DST_ADDR = server_addr

        seg = Segment()
        seg.set_seq_no(300)
        seg.set_ack_no(rcvd.get_seq_no())
        seg.setSYN()
        seg.setACK()
        seg.set_checksum(seg.count_checksum())
        self.send(seg.to_bytes(), server_addr)
        log(f"Segment SEQ={rcvd.get_seq_no()}, SYN", "Received, Ack sent")
        log(f"Segment SEQ={seg.get_seq_no()}, SYN", "Sent")

        byte, _ = self.recv()
        rcvd.from_bytes(byte)
        if (not rcvd.test_checksum()) or (not rcvd.getACK()) or (rcvd.get_ack_no() != seg.get_seq_no()):
            raise ConnectionError
        log(f"Segment SEQ={rcvd.get_ack_no()}", "Acked")

        return

    def recv_file(self):
        base = 0 # antrian ngirim data
        data_ret = b''

        log("!", f"Waiting for server to send data")
        segments = []
        while True:
            data, server_addr = self.recv() # nerima data dari server
            rcvd = Segment()
            rcvd.from_bytes(data) # unwrap the data

            if rcvd.test_checksum() and rcvd.get_seq_no() == base: # untuk ngecek valid checksumnya dan memastikan ini urutan awal

                if rcvd.isData(): # check jika segmennya bertipe data
                    ack_segment = Segment()
                    ack_segment.setACK()
                    ack_segment.set_seq_no(0)
                    ack_segment.set_ack_no(rcvd.get_seq_no())
                    ack_segment.set_checksum(ack_segment.count_checksum())
                    self.send(ack_segment.to_bytes(), server_addr)

                    log("!", f"Segment SEQ={rcvd.get_seq_no()}: Received {rcvd.isData()}, Sent {ack_segment.getACK()}")

                    data_ret += ack_segment.get_data()
                    segments.append(rcvd)
                else:
                    log("!", f"Data is empty")

                base += 1

            elif rcvd.getFIN():
                # break, will go to close_con
                log("!", "Closing Connection ...")
                seg = Segment(300, rcvd.get_seq_no())
                seg.setACK()
                seg.set_checksum(seg.count_checksum())
                self.send(seg.to_bytes(), self.DST_ADDR)
                log(f"Segment SEQ={rcvd.get_seq_no()}, FIN", "Received, Ack sent")

                time.sleep(0.1)

                seg = Segment(seg.get_seq_no())
                seg.setFIN()
                seg.set_checksum(seg.count_checksum())
                self.send(seg.to_bytes(), self.DST_ADDR)
                log(f"Segment SEQ={seg.get_seq_no()}, FIN", "Sent")

                byte, _ = self.recv()
                rcvd = Segment().from_bytes(byte)
                if (not rcvd.test_checksum()) or (not rcvd.getACK()) \
                    or (rcvd.get_ack_no() != seg.get_seq_no()):
                    raise ConnectionError
                log(f"Segment SEQ={seg.get_seq_no()}, FIN", "Acked")
                self.segments = segments
                break
            
            elif not rcvd.test_checksum() or rcvd.get_seq_no() > base:
                log("!", f"WARNING!!! Segment refused")

            elif rcvd.get_seq_no() < base:
                log("!", f"WARNING!!! Segment already received")
        return

    def save_file(self):
        fb = FileBuffer()
        fb.write(self.segments, self.FILEPATH)

if __name__=="__main__":
    if len(sys.argv) > 2:
        Client(port=int(sys.argv[1]), filepath=sys.argv[2])
    elif len(sys.argv) > 1:
        Client(port=int(sys.argv[1]))
    else :
        Client()